#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <errno.h>
#include <time.h>
#include <ctype.h>
#include <sys/param.h>
#include <sys/file.h>
#include <signal.h>

#ifdef HAVE_ZLIB
#include <zlib.h>
#endif

#include "mrmacs.h"
#include "prefs.h"
#include "config.h"
#include "conf.h"
#include "vxerror.h"
#include "tools.h"
#include "version.h"
#include "vxopt.h"
#include "exithl.h"
#include "sigwrapper.h"
#include "tmpfile.h"

#include "userdb.h"
#include "deliver.h"
#include "rexec.h"

static char *tempfile = NULL;

void cleanup (void)
{
    if (tempfile != NULL) {
	vxlog (2, "removing file `%s'.\n", tempfile);
	unlink (tempfile);
	gfree (tempfile); tempfile = NULL;
    }
}

SIGWRP(sighandle)

void sighandle (int signum)
{
    cleanup ();
    vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__,
		"%s: caught signal number %d\n",
		progname, signum);
#if 0
    /* Force default signal handling - even if BSD's signal handling is
    ** installed.
    */
    signal (signum, SIG_DFL);
#endif
    kill (getpid (), signum);
}

void usage (void)
{
    fprintf (stderr,
	     "\nUsage: %s [-f <configfile>] [-l <logfile>] [sender-addr "
	     "email-addr ...]"
	     "\n       %s -h"
	     "\n       %s -V\n"
	     "\n   -f <configfile>"
	     "\n       read configuration from file <configfile>"
	     "\n       (default: %s)"
	     "\n   -h"
	     "\n       show this help"
	     "\n   -V"
	     "\n       show the program's version number\n\n",
	     progname, progname, progname, CONFIGFILE);
    exit (EX_USAGE);
}


int main (int argc, char *argv[])
{
    FILE *fp = (FILE *) 0;
    char tmpfname[60];
    char line[1024];
    int ix, c, derrs;
    char *sender = "", *logname = (char *) 0;
/*    int ll = strlen (argv[0]);*/
    size_t sz;

    /* Settings: */
    vxopt (&argc, argv, usage);
    reset_umask();

    /* Option checking: */
    for (;;) {
	c = getopt (argc, argv, "l:v");
	if (c == -1) break;
	switch (c) {
	    case 'l':
		if (logname) {
		    fprintf (stderr, "ambigeous `-l'-option!\n");
		    break;
		}
		logname = gdupl (__FILE__, __LINE__, optarg);
		break;
	    case 'v':
		++loglevel;
		break;
	    default:
		vxerror_at (EX_USAGE, 0,
			    __FILE__, __LINE__ - 1,
			    "unknown option `%s'!\n"
			    "Use `%s -h' for help, please!\n",
			    argv[optind], progname);
	}
    }

    vxsetlog (logname);
    if (prefs.log_errors) { vxerr2log (); } else { vxerr2null (); }
    vxlog (5, "Reading configuration file ... ");
    getconf ();
    get_ug ();
    vxlog (5, "trustees ... ");
    read_trusted ();
    vxlog (5, "done.\n");

    sprintf (tmpfname, "vxdeliver.%d", getpid ());
    tempfile = gconc (__FILE__, __LINE__, prefs.tmpdir, "/", tmpfname, NULL);

    /* User database initialisieren */
    init_userdb ();

    if (argc - optind < 2) {
	vxerror_at (EX_USAGE, 0,
		    __FILE__, __LINE__ - 2,
		    "you hav to specify at least one e-mail address!\n"
		    "Use `%s -h' for help, please!\n",
		    progname);
    }

    /* Wenn ich erst hier userid und group id setze, dann kann die
    ** Konfiguration auch in einer Datei stehen, die nicht `vxuid'
    ** gehrt und fr `vxuid' gesperrt ist.
    */
    setuid (geteuid ());
    setgid (prefs.vxgid);
    setreuid (getuid (), prefs.vxuid);

    /* Argument checking: */
    sender = argv[optind++];
    if (!strcmp (sender, "ENVIRON")) {
	if (!(sender = getenv ("SENDER"))) {
	    vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
			"no sender address found!\n");
	}
    }

    /* install an exit handler `cleanup' which (should) remove(s) the
    ** temporary file created from the standard input
    */
    add_exithl ("cleanup", cleanup);

    vxlog (1, "creating message file ...");
    /* write the `stdin' contents to a temporary file */
    ifnull (fp = fvxtmp()) {
	vxlog (1, " failed.\n");
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1,
		    "fvxtmp() failed", tempfile);
    }
    sz = 0;
    while (noNULL(fgets (line, 1024, stdin))) {
	vxlog (5, ".");
	fputs (line, fp);
	sz += strlen (line);
    }
    rewind (fp);
    vxlog (1, " done.\n");
    set_msgsize (sz);
    /* Check the size of the incoming mail
    **   - size checking is generally enabled and
    **   - if delivery size checking is enabled
    */
    for (ix = optind; ix < argc; ++ix) {
	char *ra;
	/* Hiermit versuche ich einigen Schwierigkeiten, die mir der
	** `sendmail' bereitet, aus dem Weg zu gehen.
	*/
	/* So, jetzt kann der Name des Folders bestimmt werden. */
	ra = argv[ix];
	/* Remove an enclosing angle-bracket pair */
	if (*ra == '<' && ra[strlen(ra) - 1] == '>') {
	    ++ra; ra[strlen(ra) - 1] = '\0';
	}
	deliver_local (NULL, sender, ra, fp, 0);
    }
    setreuid (getuid(), getuid());
    /* The temporary file need not be deleted - as this is done automatically
    ** when it's closed.
    unlink (tempfile);
    */
    derrs = delivery_errs();
    if (prefs.errors != ERRORS_ADMIN && derrs > 0) {
	vxerror (EX_NOUSER, 0,
		 "message delivery failed for %d recipient%s!\n",
		 derrs, (derrs == 1 ? "" : "s"));
    } else if (derrs < 0) {
	derrs = - derrs;
	vxerror (EX_NOUSER, 0,
		 "there %s %d error%s during the error delivery\n",
		 (derrs == 1 ? "was" : "were"), derrs,
		 (derrs == 1 ? "" : "s"));
    }
    exit (0);
    return 0;
}
