/* dump.c
**
** database dump functions for `vxquery'
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>

#include "prefs.h"
#include "conf.h"
#include "vxerror.h"
#include "tools.h"
#include "dbif.h"
#include "version.h"
#include "vxopt.h"
#include "exithl.h"
#include "mrmacs.h"

#include "dump.h"

typedef int (*slctfunc_t) (char *lk, dbif_rec *lr, char *rk, dbif_rec *rr);
static void print_ilinks (DBIF_FILE userdb, char *domain);
static void print_xlinks (DBIF_FILE userdb, char *domain);
static void print_elinks (DBIF_FILE userdb, char *domain);
static void print_dlinks (DBIF_FILE db, char *domain);
static void print_users (DBIF_FILE db, char *domain, char *defadm, int nonstd);

static void print_admin (DBIF_FILE db, char *defadm, char *tdom);
static const char *kmv_size (size_t x);
static const char *kmv_offs (off_t x);

int dump_database (DBIF_FILE userdb)
{
    
    DBIF_CURSOR slct;
    char *defadm, *key;
    dbif_rec *rec;
    int users, maxusers;

    /* Read the `::VXMAIL::' record */
    ifnull (rec = dbif_get (userdb, "::VXMAIL::")) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "reading `::VXMAIL::'-record failed: %s\n",
		    dbif_errormsg ());
    }

    if (dbif_rectype(rec) != RT_IDENTITY) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "*** FATAL! database inconsistency ***\n");
    }

    /* write the head of the dump file */
    printf ("NEWDB \"%s\"\n", rec->folder);
    printf ("MAILHOST \"%s\"\n", rec->passwd);

    /* The default admin is required for the correct construction of the
    ** DOMAIN-sections
    */
    defadm = gdupl (__FILE__, __LINE__, rec->folder);
    dbif_free (rec);

    /* Get a list of all domains from the database */
    ifnull (slct = dbif_select (userdb, NULL, st_domain))
    {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "domain entry selection failed: %s\n",
		    dbif_errormsg ());
    }

    /* Dump the DOMAIN sections and all their (normal) user entries (i.e. no
    ** `LINKTO'-Entries; ignore LINKTO's as these may cross-refer to another
    ** domain which may be listed later.
    */
    key = dbif_first (slct, &rec);
    while (key != NULL) {
	int rt = dbif_rectype (rec);
	if (rt == RT_DOMAIN) {
	    dbif_getlim (rec, &users, &maxusers);
	    if (maxusers) {
		printf ("DOMAIN \"%s\" \"%s\" <%s %s> %d\n", key, rec->folder,
			kmv_size (rec->msg_limit), kmv_offs (rec->fld_limit),
			maxusers);
	    } else {
		printf ("DOMAIN \"%s\" \"%s\" <%s %s>\n", key, rec->folder,
			kmv_size (rec->msg_limit), kmv_offs (rec->fld_limit));
	    }
	} else {
	    printf ("NONSTD DOMAIN \"%s\" \"%s\" <%s %s>\n", key, rec->folder,
		    kmv_size (rec->msg_limit), kmv_offs (rec->fld_limit));
	}
	print_users (userdb, key, defadm, dbif_rectype(rec) == RT_NSDOMAIN);
	print_ilinks (userdb, key);
	if (rt == RT_DOMAIN) { print_admin (userdb, defadm, key); }
	free (key); dbif_free (rec);
	key = dbif_next (slct, &rec);
    }

    /* Now, dump the DOMAIN sections again but now only with their
    ** `LINKTO'-entries.
    */
    key = dbif_first (slct, &rec);
    while (key != NULL) {
	print_elinks (userdb, key);
	free (key); dbif_free (rec);
	key = dbif_next (slct, &rec);
    }
    dbif_end (slct);

    /* Now, the `LINKTO'-entries which refer to DOMAINS must be listed. */
    ifnull (slct = dbif_select (userdb, NULL, st_domlnk))
    {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "domain link selection failed: %s\n",
		    dbif_errormsg ());
    }

    /* Dump the domain links */
    key = dbif_first (slct, &rec);
    while (key != NULL) {
	printf ("DOMAIN \"%s\" LINKTO \"%s\"\n", key, rec->folder);
	free (key); dbif_free (rec);
	key = dbif_next (slct, &rec);
    }
    dbif_end (slct);

    /* Select the IP-translations and REDIRECTs from the database */
    ifnull (slct = dbif_select (userdb, NULL, st_transl))
    {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "IPxTRANS/REDIRECT selection failed: %s\n",
		    dbif_errormsg ());
    }

    /* Dump the selected entries */
    key = dbif_first (slct, &rec);
    while (key != NULL) {
	switch (dbif_rectype (rec)) {
	    case RT_IP4TRANS:
		printf ("IP4TRANS %s \"%s\"\n", key, rec->folder); break;
	    case RT_IP6TRANS:
		printf ("IP6TRANS %s \"%s\"\n", key, rec->folder); break;
	    case RT_REDIRECT:
		printf ("REDIRECT \"%s\" \"%s\" TO \"%s\"\n",
			key, rec->passwd, rec->folder);
		break;
	}
	free (key); dbif_free (rec);
	key = dbif_next (slct, &rec);
    }
    dbif_end (slct);

    /* Now, dump the `::VXTS::'-record */
    unlessnull (rec = dbif_get (userdb, "::VXTS::")) {
	if (dbif_rectype(rec) != RT_VXTS) {
	    vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
			"*** FATAL! database inconsistency ***\n");
	}
	printf ("VXTS \"%s\"\n", rec->passwd);
	dbif_free (rec);
    }

    /* Last not least, dump the VxMail system's administration entry */
    ifnull (rec = dbif_get (userdb, "::MAILSYS::")) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "Reading VxMail's system administration entry failed: %s\n",
		    dbif_errormsg());
    }
    if (dbif_rectype(rec) != RT_MAILSYS) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "*** FATAL! database inconsistency ***\n");
    }
    printf ("MAILSYS \"%s\" \"%s\"\n", rec->passwd, rec->folder);
    dbif_free (rec);
    fflush (stdout);
    return 0;
}

char *default_admin (DBIF_FILE userdb)
{
    dbif_rec *rec;
    char *res;
    ifnull (rec = dbif_get (userdb, "::VXMAIL::")) {
	return (char *) NULL;
    }
    res = gdupl (__FILE__, __LINE__, rec->folder);
    return res;
}

static __inline__ int eqnc (char *a, char *b, int lc)
{
    return (lc ? lccmp (a, b, -1) : strcmp (a, b)) == 0;
}

int dump_domain (DBIF_FILE userdb, char *domain, int rc)
{
    dbif_rec *rec;
    int rt, nu, maxu;
    char *defadm, *tdom;

    /**/
    ifnull (defadm = default_admin (userdb)) {
	vxerror_at (EX_SOFTWARE, errno, __FILE__, __LINE__,
		    "invalid database");
    }

    /* Suche zuerst mal den DOMAIN-Eintrag */
    ifnull(rec = dbif_resolve (userdb, domain, &tdom, 1)) {
	vxerror (0, 0, "domain `%s' not found!\n", domain);
	return 1;
    }

    ifnull (tdom) { tdom = gdupl (__FILE__, __LINE__, domain); }

    /* Blder Mist! Ich werde das bei allernchster Gelegenheit durch was
    ** anderes ersetzen.
    */
    rt = dbif_rectype(rec);
    if (rt != RT_DOMAIN && rt != RT_NSDOMAIN) {
	vxerror (0, 0, "domain `%s' not found!\n", tdom);
	return 1;
    }

    if (rc) {
	/* Wenn aus dem Dump die Domain neu generiert werden soll, dann mssen
	** smtliche zugehrigen Werte ausgegeben werden. Leider ist eine
	** vollstndige Generierung der Domain (einschlielich aller Verweise
	** auf die Domain und aller Benutzerverweise in die Domain) hiermit
	** NICHT mglich! Alle Benutzerverweise aus anderen Domains gehen
	** verloren.
	*/
	printf("NO DOMAIN \"%s\"\n", tdom);
	if (rt == RT_DOMAIN) {
	    dbif_getlim (rec, &nu, &maxu);
	    if (maxu) {
		printf ("DOMAIN \"%s\" \"%s\" <%s %s> %d\n", tdom, rec->folder,
			kmv_size (rec->msg_limit), kmv_offs (rec->fld_limit),
			maxu);
	    } else {
		printf ("DOMAIN \"%s\" \"%s\" <%s %s>\n", tdom, rec->folder,
			kmv_size (rec->msg_limit), kmv_offs (rec->fld_limit));
	    }
	} else {
	    printf ("NONSTD DOMAIN \"%s\" \"%s\" <%s %s>\n", tdom, rec->folder,
		    kmv_size (rec->msg_limit), kmv_offs (rec->fld_limit));
	}
    } else {
	/* Wenn keine Neuerzeugenden Direktiven gewnscht werden, dann darf
	** nur der Domainname hinter `DOMAIN' angegeben werden.
	*/
	printf ("DOMAIN \"%s\" <%s %s>\n", tdom,
		kmv_size (rec->msg_limit), kmv_offs (rec->fld_limit));
    }
    /* Zuerst werden alle regulren (nicht-LINKTO-)Eintrge ausgegeben. Da
    ** alle `LINKTO'-Eintrge mit ausgegeben werden sollen, darf der ADMIN-
    ** Eintrag hier noch nicht erscheinen. Deshalb wird als letztes Argument
    ** eine 1 (true) angegeben.
    */
    print_users (userdb, tdom, defadm, 1);

    /* Nun werden alle Verweise innerhalb der Domain angegeben. */
    print_ilinks (userdb, tdom);
    print_xlinks (userdb, tdom);

    /* Zum Schlu wird der ADMIN-Eintrag (von Hand) ausgegeben. */
    if (rt == RT_DOMAIN) { print_admin (userdb, defadm, tdom); }

    if (rc) {
	/* Na, ja - nicht ganz zum Schlu; wenn die Domain neu generiert werden,
	** dann mssen auch alle externen Verweise auf Benutzer innerhalb der
	** Domain sowie alle Verweise auf die Domain ausgegeben werden.
	*/
	print_elinks (userdb, tdom);
	print_dlinks (userdb, tdom);
    } else if (!eqnc (domain, tdom, prefs.lcdomain)) {
	/* ansonsten mu, falls ein Verweis auf diese Domain angegeben wurde,
	** wenigstens dieser Verweis mit ausgegeben werden.
	*/
	printf ("DOMAIN \"%s\" LINKTO \"%s\"\n", domain, tdom);
    }
    /* Freigabe der temporr reservierten Speicherbereiche */
    free (tdom); dbif_free (rec);

    /* Wenn die Programmausfhrung hier anlangt, dann war diese Operation
    ** erfolgreich.
    */
    return 0;
}

int dump_specials (DBIF_FILE userdb, char *sptype)
{
    char *key, *fmt;
    dbif_rec *rec;
    DBIF_CURSOR sc;
    int rtype = isNULL(sptype) ? 0
		:   eqnc (sptype, "IPTRANS", 1) ? 1
		    :   eqnc (sptype, "REDIRECT", 1) ? -1 : 0;

    if (rtype != 0) {
	/* SELECT a list of entries from the database and display it */
	sc = dbif_select (userdb, NULL, (rtype < 0 ? st_redirect : st_iptrans));
	key = dbif_first (sc, &rec);
	while (noNULL (key)) {
	    rtype = dbif_rectype (rec);
	    fmt = rtype == RT_IP4TRANS ? "IP4TRANS %s %0s\"%s\"\n"
		  :   rtype == RT_IP6TRANS ? "IP6TRANS %s %0s\"%s\"\n"
		      :   "REDIRECT \"%s\" \"%s\" TO \"%s\"\n";
	    printf (fmt, key, rec->passwd, rec->folder);
	    free (key); dbif_free (rec);
	    key = dbif_next (sc, &rec);
	}
	return 0;
    } else if (eqnc (sptype, "IDENTITY", 1)) {
	/* Display the contents of the identity record */
	rec = dbif_get (userdb, "::VXMAIL::");
	if (dbif_rectype (rec) != RT_IDENTITY) {
	    vxerror_at (0, 0, __FILE__, __LINE__ - 2,
			"database inconsistency!\n");
	    return 1;
	}
	printf ("NEWDB \"%s\"\n", rec->folder);
	printf ("MAILHOST \"%s\"\n", rec->passwd);
	dbif_free (rec);
	return 0;
    } else if (eqnc (sptype, "VXTS", 1)) {
	/* Display the VTXS password */
	rec = dbif_get (userdb, "::VXTS::");
	unlessnull (rec) {
	    if (dbif_rectype (rec) != RT_VXTS) {
		vxerror_at (0, 0, __FILE__, __LINE__ - 3,
			    "database inconsistency!\n");
		return 1;
	    }
	    printf ("VXTS \"%s\"\n", rec->passwd);
	    dbif_free (rec);
	}
	return 0;
    } else {
	/* Write an error message and exit! */
	vxerror (0, 0, "invalid type of special!\n");
	return 1;
    }
}


static void print_xilinks (DBIF_FILE userdb, char *domain, seltype sel);

/* Gib alle Links innerhalb der Domain aus, die in die gleiche Domain
** zeigen.
*/
static void print_ilinks (DBIF_FILE userdb, char *domain)
{
    print_xilinks (userdb, domain, st_ilink);
}

/* Gib alle Links innerhalb der Domain aus, die nicht in die gleiche Domain
** zeigen.
*/
static void print_xlinks (DBIF_FILE userdb, char *domain)
{
    print_xilinks (userdb, domain, st_xlink);
}

static void print_xilinks (DBIF_FILE userdb, char *domain, seltype sel)
{
    char *key;
    dbif_rec *rec;
    DBIF_CURSOR slct = dbif_select (userdb, domain, sel);
    ifnull (slct) { return; }
    key = dbif_first (slct, &rec);
    while (noNULL (key)) {
	char *p = strchr (key, '@');
	*p = '\0';
	p = strchr (rec->folder, '@') + 1;
	if (eqnc (p, domain, prefs.lcdomain)) { *--p = '\0'; }
	printf ("USER \"%s\" LINKTO \"%s\"\n", key, rec->folder);
	free (key); dbif_free (rec);
	key = dbif_next (slct, &rec);
    }
    dbif_end (slct);
}

/* Gib alle Links aus, die von dieser Domain in andere verweisen
*/
static void print_elinks (DBIF_FILE userdb, char *domain)
{
    char *key, *lastdom, *p;
    dbif_rec *rec;
    DBIF_CURSOR slct = dbif_select (userdb, domain, st_elink);
    ifnull (slct) { return; }
    lastdom = (char *) NULL;
    key = dbif_first (slct, &rec);
    while (noNULL (key)) {
	p = strrchr (key, '@'); ++p;
	ifnull (lastdom) {
	    lastdom = gdupl (__FILE__, __LINE__, p);
	    printf ("DOMAIN \"%s\"\n", p);
	} else if (!eqnc (lastdom, p, prefs.lcdomain)) {
	    gfree (lastdom); lastdom = gdupl (__FILE__, __LINE__, p);
	    printf ("DOMAIN \"%s\"\n", p);
	}
	p[-1] = '\0';
	printf ("USER \"%s\" LINKTO \"%s\"\n", key, rec->folder);
	free (key); dbif_free (rec);
	key = dbif_next (slct, &rec);
    }
    unlessnull (lastdom) { gfree (lastdom); }
    dbif_end (slct);
}

static void print_dlinks (DBIF_FILE db, char *domain)
{
    char *key;
    dbif_rec *rec;
    DBIF_CURSOR slct = dbif_select (db, domain, st_dlink);
    ifnull (slct) { return; }
    key = dbif_first (slct, &rec);
    while (noNULL (key)) {
	printf ("DOMAIN \"%s\" LINKTO \"%s\"\n", key, rec->folder);
	free (key); dbif_free (rec);
	key = dbif_next (slct, &rec);
    }
    dbif_end (slct);
}

static void print_user (char *ukey, dbif_rec *urec, int isadmin);

static void print_users (DBIF_FILE db, char *domkey, char *defadm, int nsd)
{
    char *dakey = (char *) NULL, *ukey, *p;
    dbif_rec *darec = (dbif_rec *) 0, *urec;
    DBIF_CURSOR usl;
    ifnull (usl = dbif_select (db, domkey, st_nolink))
    {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "selecting users from domain `%s' failed!\n");
    }
    for (ukey=dbif_first(usl, &urec); noNULL(ukey); ukey=dbif_next(usl, &urec))
    {
	ifnull (p = strchr (ukey, '@')) {
	    fprintf (stderr, "inconsistent entry in the user database!\n");
	} else {
	    *p = '\0';
	}
	if (!strcmp (ukey, defadm)) {
	    unlessnull (dakey) {
		fprintf (stderr, "inconsistent entry in the user database!\n");
	    } else {
		dakey = ukey;
		darec = urec;
		continue;
	    }
	}
	print_user (ukey, urec, 0);
	free (ukey); dbif_free (urec);
    }
    dbif_end (usl);
    ifnull (dakey) {
	if (!nsd) {
	    fprintf (stderr, "missing admin entry for domain `%s'!\n", domkey);
	}
	return;
    }
/*    print_user (dakey, darec, 1);*/
    free (dakey); dbif_free (darec);
}

static __inline__ void print_llto (char *type, char *key, dbif_rec *urec)
{
    char *fld = urec->folder;
    int mx = ((int) *fld++) + 1, ix;
    printf ("%s", type);
    unlessnull (key) { printf (" \"%s\"", key); }
    printf (" LLTO");
    for (ix = 0; ix < mx; ++ix) {
	printf (" \"%s\"", fld);
	fld += strlen (fld) + 1;
    }
    printf ("\n");
}

static void print_user (char *ukey, dbif_rec *urec, int isadmin)
{
    char *fn;
    switch (dbif_rectype (urec)) {
	case RT_LLINK:
	    if (isadmin) {
		print_llto ("ADMIN", NULL, urec);
	    } else {
		print_llto ("USER", ukey, urec);
	    }
	    break;
	case RT_FILTER:
	    if (!isadmin) {
		printf ("FILTER \"%s\"\n", urec->folder);
	    }
	    break;
	case RT_BSMTP:
	    if (isadmin) {
		printf ("ADMIN \"%s\" BSMTP \"%s\"\n",
			urec->passwd, urec->folder);
	    } else {
		printf ("USER \"%s\" \"%s\" <%s %s> BSMTP \"%s\"\n",
			ukey, urec->passwd, kmv_size (urec->msg_limit),
			kmv_offs (urec->fld_limit), urec->folder);
	    }
	    break;
	default:
	    if (isadmin) {
		printf ("ADMIN \"%s\"\n", urec->passwd);
	    } else {
		printf ("USER \"%s\" \"%s\" <%s %s>", ukey, urec->passwd,
			kmv_size (urec->msg_limit), kmv_offs (urec->fld_limit));
		switch (dbif_rectype(urec)) {
		    case RT_ADMIN:   printf (" ADMIN"); break;
		    case RT_INVALID: printf (" INVALID"); break;
		    default:         break;
		}
		fn = strrchr (urec->folder, '/');
		unlessnull (fn) {
		    ++fn;
		    if (strcmp(fn, ukey)) { printf(" FOLDER \"%s\"", fn); }
		}
		printf ("\n");
	    }
    }
}

static void print_admin (DBIF_FILE db, char *defadm, char *tdom)
{
    char *adm;
    dbif_rec *rec;
    int rt;
    adm = gconc (__FILE__, __LINE__, defadm, "@", tdom, NULL);
    ifnull (rec = dbif_get (db, adm)) {
	vxerror_at (EX_SOFTWARE, errno, __FILE__, __LINE__,
		    "access to database failed");
    }
    rt = dbif_rectype (rec);
    if (rt != RT_LLINK && rt != RT_ADMIN) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__,
		    "invalid record type for `%s'\n", adm);
    }
    if (rt == RT_LLINK) {
	print_llto ("ADMIN", NULL, rec);
    } else {
	printf ("ADMIN \"%s\"\n", rec->passwd);
    }
    dbif_free (rec);
    gfree (adm);
}
static const char *kmv_size (size_t x)
{
    static char res[20];
    if (x == 0) {
        strcpy (res, "NONE");
    } else if (x % 1048576 == 0) {
        sprintf (res, "%dM", x / 1048576);
    } else if (x % 1024 == 0) {
        sprintf (res, "%dK", x / 1024);
    } else {
        sprintf (res, "%d", x);
    }
    return (char *) res;
}

static const char *kmv_offs (off_t x)
{
    static char res[20];
    if (x == 0) {
        strcpy (res, "NONE");
    } else if (x % 1048576 == 0) {
        sprintf (res, "%ldM", x / 1048576);
    } else if (x % 1024 == 0) {
        sprintf (res, "%ldK", x / 1024);
    } else {
        sprintf (res, "%ld", x);
    }
    return (char *) res;
}
