/* selcc.c
**
** select conditions
**
*/

#include <string.h>

#include "prefs.h"
#include "exitcodes.h"
#include "vxerror.h"
#include "tools.h"
#include "mrmacs.h"
#include "dbif.h"
#include "selcc.h"

static __inline__ int eqnc (const char *a, const char *b, int lc)
{
    return (lc ? lccmp (a, b, -1) : strcmp (a, b)) == 0;
}

static int is_user (const char *lk, const dbif_rec *lrec,
		    const char *rk, const dbif_rec *rrec)
{
    int rt = 0;
    if (lk) {
	rt = dbif_rectype (lrec);
	if ((1 << rt) & (CC_USER|CC_NSUSER|CC_ADMIN)) {
	    return (rt == dbif_rectype (rrec)) && !strcmp (lk, rk);
	}
	if ((1 << rt) & (CC_DOMAIN|CC_NSDOMAIN)) {
	    char *p;
	    rt = dbif_rectype (rrec);
	    if (!((1 << rt) & (CC_USER|CC_ADMIN|CC_NSUSER))) { return 0; }
	    p = strchr (rk, '@');
	    if (!p) { return 0; }
	    return !strcmp (++p, lk);
	}
	return 0;
    } else {
	return ((1 << rt) & (CC_USER|CC_NSUSER|CC_ADMIN|CC_MAILSYS));
    }
}

static int udrel (const char *user, const dbif_rec *urec,
		  const char *link, const dbif_rec *lrec)
{
    if (!strcmp (user, link)) {
	return 1;
    }
    if (dbif_rectype(lrec) == RT_LINK && !strcmp (lrec->folder, user)) {
	return 1;
    }
    return 0;
}

static int ddrel (const char *dom, const dbif_rec *drec,
		  const char *ent, const dbif_rec *erec)
{
    char *dname;
    int rt;
    if (!strcmp (dom, ent)) { return 1; }
    rt = dbif_rectype(erec);
    if ((1 << rt) & CC_DENTRIES) {
	if (!(dname = strchr (ent, '@'))) { return 0; }
	++dname;
	if (!strcmp (dname, dom)) { return 1; }
	if (rt == RT_LINK) {
	    if (!(dname = strchr (erec->folder, '@'))) { return 0; }
	    ++dname;
	    if (!strcmp (dname, dom)) { return 1; }
	}
    } else if (rt == RT_DOMLINK && !strcmp (erec->folder, dom)) {
	return 1;
    }
    return 0;
}

static int is_domain (const char *lkey, const dbif_rec *lrec,
		      const char *rkey, const dbif_rec *rrec)
{
    int rt = dbif_rectype (rrec);
    return (rt == RT_DOMAIN) || (rt == RT_NSDOMAIN);
}

static int is_domlnk (const char *lkey, const dbif_rec *lrec,
		      const char *rkey, const dbif_rec *rrec)
{
    int rt = dbif_rectype (rrec);
    return (rt == RT_DOMLINK);
}

static int is_dlink (const char *lkey, const dbif_rec *lrec,
		     const char *rkey, const dbif_rec *rrec)
{
    int rt = dbif_rectype (rrec);
    if (rt != RT_DOMLINK) { return 0; }
    ifnull (lkey) { return 1; }
    return eqnc (lkey, rrec->folder, prefs.lcdomain);
}

static int is_transl (const char *lkey, const dbif_rec *lrec,
		      const char *rkey, const dbif_rec *rrec)
{
    int rt = dbif_rectype (rrec);
    return (rt == RT_IP4TRANS) || (rt == RT_IP6TRANS) || (rt == RT_REDIRECT);
}

static int is_nolink (const char *lkey, const dbif_rec *lrec,
		      const char *rkey, const dbif_rec *rrec)
{
    int rt = dbif_rectype (rrec);
    char *p;
    if (!((1 << rt) & (CC_DENTRIES - CC_LINK))) { return 0; }
    ifnull (lkey) { return 1; }
    ifnull (p = strchr (rkey, '@')) { return 0; }
    ++p;
    return eqnc (lkey, p, prefs.lcdomain);
}


static int is_link (const char *lkey, const dbif_rec *lrec,
		    const char *rkey, const dbif_rec *rrec)
{
    int rt = dbif_rectype (rrec);
    char *p;
    if (rt != RT_LINK) { return 0; }
    ifnull (lkey) { return 1; }
    ifnull (p = strchr (rkey, '@')) { return 0; }
    ++p;
    return eqnc (lkey, p, prefs.lcdomain);
}

static int is_ilink (const char *lkey, const dbif_rec *lrec,
		     const char *rkey, const dbif_rec *rrec)
{
    int rt = dbif_rectype (rrec);
    char *p, *q;
    if (rt != RT_LINK) { return 0; }
    ifnull (p = strrchr (rkey, '@')) { return 0; }
    ifnull (q = strrchr (rrec->folder, '@')) { return 0; }
    ++p; ++q;
    return eqnc (lkey, p, prefs.lcdomain) && eqnc (lkey, q, prefs.lcdomain);
}

static int is_xlink (const char *lkey, const dbif_rec *lrec,
		     const char *rkey, const dbif_rec *rrec)
{
    int rt = dbif_rectype (rrec);
    char *p, *q;
    if (rt != RT_LINK) { return 0; }
    ifnull (p = strrchr (rkey, '@')) { return 0; }
    ifnull (q = strrchr (rrec->folder, '@')) { return 0; }
    ++p; ++q;
    return eqnc (lkey, p, prefs.lcdomain) && !eqnc (lkey, q, prefs.lcdomain);
}

static int is_elink (const char *lkey, const dbif_rec *lrec,
		     const char *rkey, const dbif_rec *rrec)
{
    int rt = dbif_rectype (rrec);
    char *p, *q;
    if (rt != RT_LINK) { return 0; }
    /* rkey ist der Name des Verweises */
    ifnull (p = strrchr (rkey, '@')) { return 0; }
    /* rrec->folder ist das Verweisziel */
    ifnull (q = strrchr (rrec->folder, '@')) { return 0; }
    ++p; ++q;
    /*     domain != Domain(Verweisname) UND domain == Domain(Verweisziel) */
    return !eqnc (lkey, p, prefs.lcdomain) && eqnc (lkey, q, prefs.lcdomain);
}

static int is_iptrans (const char *lkey, const dbif_rec *lrec,
		       const char *rkey, const dbif_rec *rrec)
{
    return (CC_IPTRANS & (1 << dbif_rectype (rrec))) != 0;
}

static int is_redirect (const char *lkey, const dbif_rec *lrec,
		       const char *rkey, const dbif_rec *rrec)
{
    return (CC_REDIRECT & (1 << dbif_rectype (rrec))) != 0;
}

selcc_t selcc (seltype sel)
{
    switch (sel) {
	case st_user: return is_user;
	case st_udel: return udrel;
	case st_ddel: return ddrel;
	case st_domain: return is_domain;
	case st_domlnk: return is_domlnk;
	case st_dlink: return is_dlink;
	case st_transl: return is_transl;
	case st_nolink: return is_nolink;
	case st_link: return is_link;
	case st_ilink: return is_ilink;
	case st_xlink: return is_xlink;
	case st_elink: return is_elink;
	case st_iptrans: return is_iptrans;
	case st_redirect: return is_redirect;
	default:
	    vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__,
			"invalid type of database selection\n");
	    return (selcc_t) 0;
    }
}
